@@ -343,7 +343,7 @@ dhcp_outputipudp(const struct interface *ifp, const struct in_addr *src,
 		udp.uh_dport = htons(BOOTPS);
 	} else {
 		if ((bootp->ciaddr == INADDR_ANY &&
-			bootp->flags & BROADCAST_FLAG) ||
+			ntohs(bootp->flags) & BROADCAST_FLAG) ||
 		    type == DHCP_NAK)
 			ip.ip_dst.s_addr = INADDR_BROADCAST;
 		else
@@ -502,8 +502,8 @@ dhcp_plugin_validateaddr(struct ctx *ctx, const struct in_addr *addr)
 }
 
 static struct in_addr
-dhcp_plugin_findaddr(struct ctx *ctx, char *hostname, const struct bootp *bootp,
-    size_t len)
+dhcp_plugin_findaddr(struct ctx *ctx, char *hostname, uint32_t *ltime,
+    const struct bootp *bootp, size_t len)
 {
 	struct plugin *p;
 	struct sockaddr_in sin = {
@@ -517,6 +517,9 @@ dhcp_plugin_findaddr(struct ctx *ctx, char *hostname, const struct bootp *bootp,
 	const char *hname = NULL;
 	size_t l;
 
+	/* We might not get a least time so set to zero for a default. */
+	*ltime = 0;
+
 	PLUGIN_FOREACH(ctx, p)
 	{
 		if (p->p_lookup_hostname == NULL)
@@ -547,8 +550,8 @@ dhcp_plugin_findaddr(struct ctx *ctx, char *hostname, const struct bootp *bootp,
 	{
 		if (p->p_lookup_addr == NULL)
 			continue;
-		err = p->p_lookup_addr(p, (struct sockaddr *)&sin, hname, bootp,
-		    len);
+		err = p->p_lookup_addr(p, (struct sockaddr *)&sin, ltime, hname,
+		    bootp, len);
 		if (err == -1) {
 			if (errno != ESRCH && errno != ENOSYS)
 				logerr("plugin %s", p->p_name);
@@ -572,7 +575,7 @@ dhcp_findpool(const struct interface *ifp, in_addr_t addr, bool strict)
 	size_t npools;
 
 	for (pool = ifp->if_pools, npools = ifp->if_npools; npools != 0;
-	     pool++, npools--) {
+	    pool++, npools--) {
 		if (ntohl(addr) >= ntohl(pool->dp_from.s_addr) &&
 		    ntohl(addr) <= ntohl(pool->dp_to.s_addr))
 			return pool;
@@ -782,27 +785,27 @@ dhcp_addoptions(struct bootp *bootp, uint8_t **p, const uint8_t *e,
     const struct dhcp_pool *pool, const char *msg, const struct bootp *req,
     size_t reqlen)
 {
-	const uint8_t *opt;
-
 	DHCP_PUT_B(p, e, DHO_MESSAGETYPE, type);
 	DHCP_PUT_U32(p, e, DHO_SERVERID, pool->dp_addr.s_addr);
 	if (type == DHCP_OFFER || type == DHCP_ACK) {
-		uint32_t u32, lease_time = ctx->dhcp_lease_time;
+		struct timespec tv;
+		uint32_t u32, lease_time, renew_time, rebind_time;
 		struct plugin *plug;
 		int n;
 
-		opt = dhcp_findoption(req, reqlen, DHO_LEASETIME);
-		if (opt != NULL && opt[0] == sizeof(u32)) {
-			memcpy(&u32, opt + 1, sizeof(u32));
-			u32 = ntohl(u32);
-			if (u32 < lease_time)
-				lease_time = u32;
-		}
+		if (timespecisset(&lease->dl_leased)) {
+			timespecsub(&lease->dl_expires, &lease->dl_leased, &tv);
+			lease_time = (uint32_t)tv.tv_sec;
+			renew_time = (uint32_t)(lease_time * T1);
+			rebind_time = (uint32_t)(lease_time * T2);
+		} else
+			lease_time = renew_time = rebind_time = INFINITE_LIFETIME;
+
 		u32 = htonl(lease_time);
 		DHCP_PUT_U32(p, e, DHO_LEASETIME, u32);
-		u32 = htonl((uint32_t)(lease_time * T1));
+		u32 = htonl(renew_time);
 		DHCP_PUT_U32(p, e, DHO_RENEWALTIME, u32);
-		u32 = htonl((uint32_t)(lease_time * T2));
+		u32 = htonl(rebind_time);
 		DHCP_PUT_U32(p, e, DHO_REBINDTIME, u32);
 
 		n = dhcp_addhostname(p, e, lease, req, reqlen);
@@ -850,6 +853,7 @@ dhcp_output(const struct interface *ifp, const struct dhcp_lease *lease,
 
 	memset(bootp, 0, sizeof(*bootp));
 	bootp->op = BOOTREPLY;
+	bootp->flags = htons(ntohs(req->flags) & BROADCAST_FLAG);
 	bootp->htype = req->htype;
 	bootp->hlen = req->hlen;
 	if (req->hlen > 0)
@@ -988,17 +992,23 @@ dhcp_set_expire_timeout(struct dhcp_ctx *ctx)
 
 static void
 dhcp_lease_settime(struct dhcp_ctx *ctx, struct dhcp_lease *lease,
-    struct timespec *now, const struct bootp *bootp, size_t len)
+    struct timespec *now, uint32_t lease_time, const struct bootp *bootp,
+    size_t len)
 {
 	const uint8_t *opt;
-	uint32_t u32, lease_time = ctx->dhcp_lease_time;
 
-	opt = dhcp_findoption(bootp, len, DHO_LEASETIME);
-	if (opt != NULL && opt[0] == sizeof(u32)) {
-		memcpy(&u32, opt + 1, sizeof(u32));
-		u32 = ntohl(u32);
-		if (u32 < lease_time)
-			lease_time = u32;
+	if (lease_time == 0) {
+		uint32_t u32;
+
+		lease_time = ctx->dhcp_lease_time;
+
+		opt = dhcp_findoption(bootp, len, DHO_LEASETIME);
+		if (opt != NULL && opt[0] == sizeof(u32)) {
+			memcpy(&u32, opt + 1, sizeof(u32));
+			u32 = ntohl(u32);
+			if (u32 < lease_time)
+				lease_time = u32;
+		}
 	}
 
 	if (lease->dl_in_expire_tree) {
@@ -1095,6 +1105,7 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 	uint8_t type, clientid[DHCP_CLIENTID_LEN + 1], fqdn_flags;
 	struct in_addr addr = { .s_addr = INADDR_ANY };
 	char phostname[DHCP_HOSTNAME_LEN] = { '\0' };
+	uint32_t pltime = 0;
 	struct in_addr paddr = { .s_addr = INADDR_ANY };
 	struct dhcp_lease *lease = NULL, *wanted = NULL;
 	char clid_buf[sizeof(clientid) * 3];
@@ -1153,8 +1164,10 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 		clientid[0] = bootp->hlen + 1;
 		clientid[1] = bootp->htype;
 		memcpy(clientid + 2, bootp->chaddr, bootp->hlen);
-	} else
+	} else {
 		clientid[0] = '\0';
+		clientid[1] = '\0'; /* silences maybe uninitialiased warning */
+	}
 
 	clid = hwaddr_ntoa(clientid + 1, clientid[0], clid_buf,
 	    sizeof(clid_buf));
@@ -1170,7 +1183,7 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 			return;
 		if (lease->dl_addr.s_addr == INADDR_ANY) {
 			lease->dl_addr = dhcp_plugin_findaddr(ifp->if_ctx,
-			    phostname, bootp, len);
+			    phostname, &pltime, bootp, len);
 			if (lease->dl_addr.s_addr != INADDR_ANY)
 				lease->dl_flags |= DL_PLUGIN_ADDRESS;
 		}
@@ -1220,8 +1233,9 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 		return;
 	}
 
-	logdebugx("%s: %s 0x%x %s: %s", ifp->if_name, flags ? "cont" : "recv",
-	    bootp->xid, dhcp_message_type(type), clid);
+	logdebugx("%s: %s 0x%x %s%s: %s", ifp->if_name, flags ? "cont" : "recv",
+	    bootp->xid, dhcp_message_type(type),
+	    ntohs(bootp->flags) & BROADCAST_FLAG ? "(bcast)" : "", clid);
 
 	lease = dhcp_newlease(ctx, clientid);
 	if (lease == NULL)
@@ -1260,8 +1274,8 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 		}
 		/* FALLTHROUGH */
 	case DHCP_REQUEST:
-		paddr = dhcp_plugin_findaddr(ifp->if_ctx, phostname, bootp,
-		    len);
+		paddr = dhcp_plugin_findaddr(ifp->if_ctx, phostname, &pltime,
+		    bootp, len);
 		if (paddr.s_addr == INADDR_ANY)
 			break;
 		wanted = dhcp_lease_findaddr(ctx, &paddr);
@@ -1365,7 +1379,7 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 			expires = lease->dl_expires;
 		else
 			timespecclear(&expires);
-		dhcp_lease_settime(ctx, lease, &now, bootp, len);
+		dhcp_lease_settime(ctx, lease, &now, pltime, bootp, len);
 		if (wanted != lease) {
 			if (dhcp_lease_insertaddr(ctx, lease) == -1) {
 				logerr("%s: dhcp_lease_insertaddr", __func__);
@@ -1417,7 +1431,7 @@ dhcp_handlebootp(struct interface *ifp, struct bootp *bootp, size_t len,
 			lease->dl_flags |= DL_PLUGIN_ADDRESS;
 		else
 			lease->dl_flags &= ~DL_PLUGIN_ADDRESS;
-		dhcp_lease_settime(ctx, lease, &now, bootp, len);
+		dhcp_lease_settime(ctx, lease, &now, pltime, bootp, len);
 		if (wanted != lease) {
 			if (dhcp_lease_insertaddr(ctx, lease) == -1) {
 				logerr("%s: dhcp_lease_insertaddr", __func__);
@@ -1536,16 +1550,16 @@ out:
 		} else {
 			memset(lease->dl_hostname, '\0',
 			    sizeof(lease->dl_hostname));
-			if (opt != NULL && opt[0] >= 3) {
+			if (opt != NULL && opt[0] >= 4) {
 				if (fqdn_flags & FQDN_E) {
 					if (decode_rfc1035(lease->dl_hostname,
 						sizeof(lease->dl_hostname),
-						opt + 3, opt[0] - 3) == -1)
+						opt + 4, opt[0] - 3) == -1)
 						memset(lease->dl_hostname, '\0',
 						    sizeof(lease->dl_hostname));
 				} else {
 					// hostname field is bigger than 255
-					memcpy(lease->dl_hostname, opt + 3,
+					memcpy(lease->dl_hostname, opt + 4,
 					    opt[0] - 3);
 				}
 			} else {
@@ -1557,7 +1571,11 @@ out:
 			if (lease->dl_hostname[0] != '\0')
 				lease->dl_flags |= DL_HOSTNAME;
 		}
-		dhcp_commit_lease(ctx, lease, bootp, len);
+		if (lease->dl_flags & DL_HOSTNAME) {
+			sanitize_rfc1035(lease->dl_hostname);
+			if (lease->dl_hostname[0] == '\0')
+				lease->dl_flags &= ~DL_HOSTNAME;
+		}
 		break;
 	}
 
@@ -1567,6 +1585,10 @@ out:
 	case DHCP_ACK:
 	case DHCP_NAK:
 		dhcp_output(ifp, lease, type, msg, bootp, len);
+		/* Commit the lease after sending it as it might
+		 * be a slow operation depending on what the plugins do. */
+		if (type != DHCP_NAK)
+			dhcp_commit_lease(ctx, lease, bootp, len);
 		break;
 	}
 }
